#include <linux/input.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <pthread.h>
#include <signal.h>
#include <poll.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <stdio.h>
#include "tbwheel.h"

static int pp[2];
static int modifier=1;

static char *devlist(void)
{
	FILE *fp;
	char *list;
	char *p;
	char *item;
	char *name;
	char *handler;
	char *dev;
	char *evdev;
	char *mdev;
	char *save;
	char bfr[1024];

	if(!(list=malloc(65536)))return NULL;
	if(!(fp=fopen(DEVICES,"r")))
	{
		free(list);
		return NULL;
	}

	p=list;
	*p=0;
	name=NULL;
	handler=NULL;

	while(fgets(bfr,sizeof(bfr),fp))
	{
		item=strtok_r(bfr," \n",&save);
		if(!item)
		{
			if(name&&handler)
			{
				evdev=NULL;
				mdev=NULL;
				dev=strtok_r(handler," ",&save);
				while(dev)
				{
					if(!strncmp(dev,"mouse",5))mdev=dev;
					else if(!strncmp(dev,"event",5))
						evdev=dev;
					dev=strtok_r(NULL," ",&save);
				}
				if(mdev&&evdev)
					p+=sprintf(p,"/dev/input/%s %s\n",
						evdev,name);
			}
			if(name)free(name);
			if(handler)free(handler);
			name=NULL;
			handler=NULL;
		}
		else if(!strcmp(item,"N:"))
		{
			if(name)free(name);
			name=strtok_r(NULL,"\"",&save);
			name=strtok_r(NULL,"\"",&save);
			if(name)name=strdup(name);
		}
		else if(!strcmp(item,"H:"))
		{
			if(handler)free(handler);
			handler=strtok_r(NULL,"=\n",&save);
			handler=strtok_r(NULL,"=\n",&save);
			if(handler)handler=strdup(handler);
		}
	}

	fclose(fp);
	if(name)free(name);
	if(handler)free(handler);
	return list;
}

static char *kbdlist(void)
{
	int f;
	FILE *fp;
	char *item;
	char *save;
	char *list;
	char *p;
	char *handler;
	char *dev;
	char *kdev;
	char *evdev;
	char bfr[1024];

	if(!(list=malloc(65536)))return NULL;
	if(!(fp=fopen(DEVICES,"r")))
	{
		free(list);
		return NULL;
	}

	f=0;
	p=list;
	*p=0;
	handler=NULL;

	while(fgets(bfr,sizeof(bfr),fp))
	{
		item=strtok_r(bfr," \n",&save);
		if(!item)
		{
			if(f&&handler)
			{
				evdev=NULL;
				kdev=NULL;
				dev=strtok_r(handler," ",&save);
				while(dev)
				{
					if(!strcmp(dev,"kbd"))kdev=dev;
					else if(!strncmp(dev,"event",5))
						evdev=dev;
					dev=strtok_r(NULL," ",&save);
				}
				if(kdev&&evdev)
					p+=sprintf(p,"/dev/input/%s\n",evdev);
			}
			f=0;
			if(handler)free(handler);
			handler=NULL;
		}
		else if(!strcmp(item,"B:"))
		{
			item=strtok_r(NULL,"=\n",&save);
			if(!strcmp(item,"LED"))f=1;
		}
		else if(!strcmp(item,"H:"))
		{
			if(handler)free(handler);
			handler=strtok_r(NULL,"=\n",&save);
			handler=strtok_r(NULL,"=\n",&save);
			if(handler)handler=strdup(handler);
		}
	}

	fclose(fp);
	if(handler)free(handler);
	return list;
}

static void *handler(void *dev)
{
	int i;
	int pc;
	int mfd;
	int state;
	int lastate;
	int rastate;
	int lsstate;
	int rsstate;
	int delta;
	char *list;
	char *item;
	char *save;
	struct input_event e;
	struct pollfd p[9];
	unsigned char msg[4];

	if((p[0].fd=open(dev,O_RDWR))==-1)goto dummy;

	if(ioctl(p[0].fd,EVIOCGRAB,(void *)1))
	{
		close(p[0].fd);
		goto dummy;
	}

	if((mfd=open(OUTPIPE,O_RDWR))==-1)
	{
		close(p[0].fd);
		goto dummy;
	}

	if(fcntl(mfd,F_SETFL,O_NONBLOCK))
	{
		close(mfd);
		close(p[0].fd);
		goto dummy;
	}

	p[0].events=POLLIN;

	p[1].fd=pp[0];
	p[1].events=POLLIN;

	pc=2;

	if((list=kbdlist()))
	{
		item=strtok_r(list,"\n",&save);
		while(item&&pc<10)
		{
			if((p[pc].fd=open(item,O_RDONLY))!=-1)
				p[pc++].events=POLLIN;
			item=strtok_r(NULL,"\n",&save);
		}

		free(list);
	}

	state=0;
	lastate=0;
	rastate=0;
	lsstate=0;
	rsstate=0;
	delta=0;

	while(1)
	{
		if(poll(p,pc,-1)<=0)continue;

		if(p[0].revents&POLLIN)
		    if(read(p[0].fd,&e,sizeof(e))==sizeof(e))switch(e.type)
		{
		case EV_REL:
			switch(e.code)
			{
			case REL_X:
				if(modifier)
				{
					if(lsstate||rsstate)break;
				}
				else if(lastate||rastate)break;
				msg[0]=0x08|(e.value<0?0x10:0x00)|(state&0x07);
				msg[1]=e.value&0xff;
				msg[2]=0x00;
				msg[3]=0x00;
				write(mfd,msg,4);
				break;

			case REL_Y:
				if(modifier)
				{
					if(lsstate||rsstate)goto emu;
				}
				else if(lastate||rastate)
				{
emu:					delta+=e.value;
					if(delta>=10)
					{
						if(delta>=20)delta=19;
						e.value=1;
						delta-=10;
					}
					else if(delta<=-10)
					{
						if(delta<=-20)delta=-19;
						e.value=-1;
						delta+=10;
					}
					else break;
					msg[0]=0x08|(state&0x07);
					msg[1]=0x00;
					msg[2]=0x00;
					msg[3]=(e.value<0?0xff:0x01);
					write(mfd,msg,4);
					break;
				}
				msg[0]=0x08|(e.value>0?0x20:0x00)|(state&0x07);
				msg[1]=0x00;
				msg[2]=(-e.value)&0xff;
				msg[3]=0x00;
				write(mfd,msg,4);
				break;

			case REL_WHEEL:
				msg[0]=0x08|(state&0x07);
				msg[1]=0x00;
				msg[2]=0x00;
				msg[3]=(e.value>0?0xf8:0x00)|((-e.value)&0x07);
				write(mfd,msg,4);
				break;
			}
			break;

		case EV_KEY:
			switch(e.code)
			{
			case BTN_LEFT:
				if(e.value)state|=0x01;
				else state&=~0x01;
				msg[0]=0x08|(state&0x07);
				msg[1]=0x00;
				msg[2]=0x00;
				msg[3]=0x00;
				write(mfd,msg,4);
				break;

			case BTN_RIGHT:
				if(e.value)state|=0x02;
				else state&=~0x02;
				msg[0]=0x08|(state&0x07);
				msg[1]=0x00;
				msg[2]=0x00;
				msg[3]=0x00;
				write(mfd,msg,4);
				break;

			case BTN_MIDDLE:
				if(e.value)state|=0x04;
				else state&=~0x04;
				msg[0]=0x08|(state&0x07);
				msg[1]=0x00;
				msg[2]=0x00;
				msg[3]=0x00;
				write(mfd,msg,4);
				break;
			}
			break;
		}

		if(p[1].revents&POLLIN)
		{
			read(p[1].fd,msg,1);
			for(i=0;i<pc;i++)if(i!=1)close(p[i].fd);
			pthread_exit(NULL);
		}

		for(i=2;i<pc;i++)if(p[i].revents&POLLIN)
		    if(read(p[i].fd,&e,sizeof(e))==sizeof(e))switch(e.type)
		{
		case EV_KEY:
			switch(e.code)
			{
			case KEY_LEFTSHIFT:
				if(e.value)lsstate|=(1<<i);
				else lsstate &=~(1<<i);
				break;

			case KEY_RIGHTSHIFT:
				if(e.value)rsstate|=(1<<i);
				else rsstate &=~(1<<i);
				break;

			case KEY_LEFTALT:
				if(e.value)lastate|=(1<<i);
				else lastate &=~(1<<i);
				break;

			case KEY_RIGHTALT:
				if(e.value)rastate|=(1<<i);
				else rastate &=~(1<<i);
				break;
			}
			break;
		}
	}

dummy:	p[0].fd=pp[0];
	p[0].events=POLLIN;

	while(1)
	{
		if(poll(p,1,-1)<=0)continue;

		if(p[0].revents&POLLIN)
		{
			read(p[0].fd,msg,1);
			pthread_exit(NULL);
		}
	}
}

static void doterm(int unused)
{
	unlink(CONTROL);
	unlink(PIDFILE);
	exit(0);
}

int main(int argc,char *argv[])
{
	int c;
	int s;
	int i;
	int fifoonly;
	int disabled;
	int foreground;
	char *list;
	char *item;
	char *dev;
	char *name;
	char *save;
	char *mousedev;
	FILE *fp;
	pthread_t h;
	mode_t m;
	struct stat stb;
	struct sockaddr_un a;
	struct pollfd p;
	char msg[130];

	disabled=0;
	fifoonly=0;
	foreground=0;
	mousedev=NULL;

	while((c=getopt(argc,argv,"qm:pfa"))!=-1)switch(c)
	{
	case 'q':
		list=devlist();
		if(!list)return 0;
		if(!*list)
		{
			free(list);
			return 0;
		}
		item=strtok_r(list,"\n",&save);
		while(item)
		{
			while(*item&&*item!=' ')item++;
			if(*item)item++;
			if(*item)printf("%s\n",item);
			item=strtok_r(NULL,"\n",&save);
		}
		free(list);
		return 0;

	case 'm':
		if(mousedev)free(mousedev);
		mousedev=strdup(optarg);
		break;

	case 'p':
		fifoonly=1;
		break;

	case 'f':
		foreground=1;
		break;

	case 'a':
		modifier=0;
		break;

	default:return 1;
	}

	if(getuid())
	{
		fprintf(stderr,"not root\n");
		return 1;
	}

	if(stat(OUTPIPE,&stb))
	{
		m=umask(0077);
		if(mknod(OUTPIPE,S_IFIFO|0600,0))
		{
			perror("mknod");
			return 1;
		}
		umask(m);
	}
	else if(!S_ISFIFO(stb.st_mode))
	{
		fprintf(stderr,"%s: not a fifo\n",OUTPIPE);
		return 1;
	}
	else if((stb.st_mode&0777)!=0600)
	{
		if(chmod(OUTPIPE,0600))
		{
			perror("chmod");
			return 1;
		}
	}

	if(fifoonly)return 0;

	if(pipe(pp))
	{
		perror("pipe");
		return 1;
	}

	if((s=socket(PF_UNIX,SOCK_STREAM,0))==-1)
	{
		perror("socket");
		return 1;
	}

	memset(&a,0,sizeof(a));
	a.sun_family=AF_UNIX;
	strcpy(a.sun_path,CONTROL);

	if(!stat(CONTROL,&stb))
	{
		if(!S_ISSOCK(stb.st_mode))
		{
			fprintf(stderr,"%s: not a socket\n",CONTROL);
			return 1;
		}
		unlink(CONTROL);
	}

	m=umask(0000);
	if(bind(s,(struct sockaddr *)(&a),sizeof(a)))
	{
		perror("bind");
		return 1;
	}
	umask(m);

	if(listen(s,5))
	{
		perror("listen");
		return 1;
	}

	if(mlockall(MCL_CURRENT|MCL_FUTURE))
	{
		perror("mlockall");
		return 1;
	}

	if(!foreground)
	{
		if(daemon(0,0))
		{
			perror("daemon");
			return 1;
		}
	}

	if((fp=fopen(PIDFILE,"w")))
	{
		fprintf(fp,"%d\n",getpid());
		fclose(fp);
	}

	signal(SIGHUP,doterm);
	signal(SIGINT,doterm);
	signal(SIGTERM,doterm);
	signal(SIGQUIT,doterm);
	signal(SIGSTOP,doterm);
	signal(SIGTSTP,doterm);
	signal(SIGTTIN,doterm);
	signal(SIGTTOU,doterm);

	while(1)
	{
		dev=NULL;

		if(mousedev&&!disabled)if((list=devlist()))
		{
			dev=strtok_r(list,"\n",&save);
			while(dev)
			{
				for(name=dev;*name&&*name!=' ';name++);
				if(*name)*name++=0;
				if(*dev&&*name)
				{
					if(!strcmp(name,mousedev))
					{
						dev=strdup(dev);
						break;
					}
				}
				dev=strtok_r(NULL,"\n",&save);
			}
			free(list);
		}

		if(dev)if(pthread_create(&h,NULL,handler,dev))
		{
			free(dev);
			dev=NULL;
		}

netloop:	while((p.fd=accept(s,NULL,NULL))==-1);
		p.events=POLLIN;

		if(poll(&p,1,100)!=1)goto netout;
		if(!(p.revents&POLLIN))goto netout;
		if(read(p.fd,msg,1)!=1)goto netout;

		switch(*msg)
		{
		case CMD_STATUS:
			if(dev)*msg=1;
			else *msg=0;
			write(p.fd,msg,1);
			goto netout;

		case CMD_DISABLE:
			disabled=1;
			*msg=1;
			write(p.fd,msg,1);
			goto repeat;

		case CMD_ENABLE:
			disabled=0;
			*msg=1;
			write(p.fd,msg,1);
			goto repeat;

		case CMD_DEVICE:
			if(poll(&p,1,100)!=1)goto netout;
			if(!(p.revents&POLLIN))goto netout;
			if(read(p.fd,msg+1,1)!=1)goto netout;
			if(msg[1]<0)goto netout;
			if(msg[1])if(read(p.fd,msg+2,msg[1])!=msg[1])
				goto netout;
			c=msg[1];
			c+=2;
			msg[c]=0;
			*msg=1;
			for(i=2;i<c;i++)if(msg[i]<0x20||msg[i]>0x7e)*msg=0;
			if(*msg)
			{
				if(mousedev)free(mousedev);
				mousedev=strdup(msg+2);
			}
			write(p.fd,msg,1);
			goto repeat;

		case CMD_USEALT:
			modifier=0;
			*msg=1;
			write(p.fd,msg,1);
			goto repeat;

		case CMD_USESHIFT:
			modifier=1;
			*msg=1;
			write(p.fd,msg,1);
			goto repeat;

		}

netout:		close(p.fd);
		goto netloop;

repeat:		close(p.fd);

		if(dev)
		{
			write(pp[1],msg,1);
			pthread_join(h,NULL);
		}
	}
}
